/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 2011 Oracle and/or its affiliates. All rights reserved.
 *
 * Oracle and Java are registered trademarks of Oracle and/or its affiliates.
 * Other names may be trademarks of their respective owners.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 *
 * Portions Copyrighted 2011 Sun Microsystems, Inc.
 */
package org.netbeans.modules.plsql.hyperlink.actions;

import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;
import javax.swing.text.Document;
import org.netbeans.api.project.FileOwnerQuery;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataObject;
import org.openide.util.Exceptions;

/**
 *
 * @author csamlk
 */
public class ShowDbObjectDialog extends javax.swing.JDialog {
   private boolean isCancelled = true;
   private boolean enterPressed = false;
   private boolean escPressed = false;

    /** Creates new form ShowDbObjectDialog */
    public ShowDbObjectDialog(java.awt.Frame parent, Project project, String defaultVal, boolean modal) {
      super(parent, modal);
      try {
         initComponents();
         //Make the editor pane a PL/SQL editor pane that supports code completion
         //showObject.setEditorKit(showObject.getEditorKitForContentType("text/x-plsql"));
         //showObject.setEditorKitForContentType("text/x-plsql", showObject.getEditorKit());
         //showObject.setContentType("text/x-plsql");
         File tmpFile = File.createTempFile("obj", ".tdb");
         tmpFile.deleteOnExit();
         FileObject tmpObj = FileUtil.createData(FileUtil.normalizeFile(tmpFile));
         FileOwnerQuery.markExternalOwner(tmpObj, project, FileOwnerQuery.EXTERNAL_ALGORITHM_TRANSIENT);
         DataObject data = DataObject.find(tmpObj);
         showObject.getDocument().putProperty(Document.StreamDescriptionProperty, data);
         showObject.setText(defaultVal);
         showObject.setSelectionStart(0);
         showObject.setSelectionEnd(defaultVal.length());
         enterPressed = false;
      } catch (IOException ex) {
         Exceptions.printStackTrace(ex);
      }
   }
  public boolean isCancelled() {
      return isCancelled;
   }

   public String getInputText() {
      return showObject.getText();
   }
   public void setLable(String lable) {
      this.lable.setText(lable);
   }
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
   // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
   private void initComponents() {

      jScrollPane1 = new javax.swing.JScrollPane();
      showObject = new javax.swing.JEditorPane();
      lable = new javax.swing.JLabel();
      okButton = new javax.swing.JButton();
      cancelBtn = new javax.swing.JButton();

      setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
      setIconImage(null);

      jScrollPane1.setVerticalScrollBarPolicy(javax.swing.ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER);

      //showObject.setEditorKit(showObject.getEditorKitForContentType("text/x-plsql"));
      showObject.setEditorKitForContentType("text/x-plsql", showObject.getEditorKitForContentType("text/x-plsql"));
      showObject.setContentType("text/x-plsql");
      showObject.addKeyListener(new java.awt.event.KeyAdapter() {
         public void keyPressed(java.awt.event.KeyEvent evt) {
            showObjectKeyPressed(evt);
         }
         public void keyReleased(java.awt.event.KeyEvent evt) {
            showObjectKeyReleased(evt);
         }
      });
      jScrollPane1.setViewportView(showObject);

      lable.setText(org.openide.util.NbBundle.getMessage(ShowDbObjectDialog.class, "ShowDbObjectDialog.lable.text")); // NOI18N

      okButton.setText(org.openide.util.NbBundle.getMessage(ShowDbObjectDialog.class, "ShowDbObjectDialog.okButton.text")); // NOI18N
      okButton.addActionListener(new java.awt.event.ActionListener() {
         public void actionPerformed(java.awt.event.ActionEvent evt) {
            okButtonActionPerformed(evt);
         }
      });

      cancelBtn.setText(org.openide.util.NbBundle.getMessage(ShowDbObjectDialog.class, "ShowDbObjectDialog.cancelBtn.text")); // NOI18N
      cancelBtn.addActionListener(new java.awt.event.ActionListener() {
         public void actionPerformed(java.awt.event.ActionEvent evt) {
            cancelBtnActionPerformed(evt);
         }
      });

      javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
      getContentPane().setLayout(layout);
      layout.setHorizontalGroup(
         layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
         .addGroup(layout.createSequentialGroup()
            .addGap(37, 37, 37)
            .addComponent(lable, javax.swing.GroupLayout.PREFERRED_SIZE, 269, javax.swing.GroupLayout.PREFERRED_SIZE)
            .addContainerGap(63, Short.MAX_VALUE))
         .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
            .addContainerGap(37, Short.MAX_VALUE)
            .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
               .addComponent(jScrollPane1, javax.swing.GroupLayout.PREFERRED_SIZE, 322, javax.swing.GroupLayout.PREFERRED_SIZE)
               .addGroup(layout.createSequentialGroup()
                  .addComponent(okButton)
                  .addGap(18, 18, 18)
                  .addComponent(cancelBtn)))
            .addContainerGap())
      );
      layout.setVerticalGroup(
         layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
         .addGroup(layout.createSequentialGroup()
            .addGap(4, 4, 4)
            .addComponent(lable)
            .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
            .addComponent(jScrollPane1, javax.swing.GroupLayout.PREFERRED_SIZE, 20, javax.swing.GroupLayout.PREFERRED_SIZE)
            .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
            .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
               .addComponent(cancelBtn)
               .addComponent(okButton))
            .addContainerGap(29, Short.MAX_VALUE))
      );

      pack();
   }// </editor-fold>//GEN-END:initComponents

    private void okButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_okButtonActionPerformed
       String inputText = showObject.getText();
       if(inputText!=null) {
          inputText=inputText.replaceAll("\r", "").replaceAll("\n", "").trim();
          showObject.setText(inputText);
       }
       isCancelled = false;
       this.setVisible(false);
       this.dispose();
    }//GEN-LAST:event_okButtonActionPerformed

    private void cancelBtnActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_cancelBtnActionPerformed
       this.setVisible(false);
       this.dispose();
    }//GEN-LAST:event_cancelBtnActionPerformed

    private void showObjectKeyPressed(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_showObjectKeyPressed
       //Workaround to make TAB key navigate to the Ok button
       if (evt.getKeyCode() == KeyEvent.VK_TAB) {
           evt.consume();
           okButton.requestFocus();
       }
       if (evt.getKeyCode() == KeyEvent.VK_ENTER) {
          enterPressed=true;
       }
       if (evt.getKeyCode() == KeyEvent.VK_ESCAPE) {
          escPressed=true;
       }
    }//GEN-LAST:event_showObjectKeyPressed

    private void showObjectKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_showObjectKeyReleased
       if (enterPressed && evt.getKeyCode() == KeyEvent.VK_ENTER) {
           evt.consume();
           okButton.requestFocus();
           okButtonActionPerformed(null);
       }
       if (escPressed && evt.getKeyCode() == KeyEvent.VK_ESCAPE) {
           evt.consume();
           cancelBtn.requestFocus();
           cancelBtnActionPerformed(null);
       }
    }//GEN-LAST:event_showObjectKeyReleased

   // Variables declaration - do not modify//GEN-BEGIN:variables
   private javax.swing.JButton cancelBtn;
   private javax.swing.JScrollPane jScrollPane1;
   private javax.swing.JLabel lable;
   private javax.swing.JButton okButton;
   private javax.swing.JEditorPane showObject;
   // End of variables declaration//GEN-END:variables

}
